@@ -14,16 +14,16 @@ import requests
 from tornado import ioloop
 
 from .contexts.default.context import DefaultContext
-from .lib.context import BaseContext, autodiscover_contexts
+from .lib.context import BeakerContext, autodiscover_contexts
 from .lib.jupyter_kernel_proxy import (KERNEL_SOCKETS, KERNEL_SOCKETS_NAMES,
                                        InterceptionFilter, JupyterMessage,
                                        KernelProxyManager)
 from .lib.utils import (message_handler, LogMessageEncoder, magic,
-                        handle_message, get_socket, server_url, server_token)
+                        handle_message, get_socket, execution_context, parent_message_context, get_parent_message)
 
 if TYPE_CHECKING:
-    from .lib.agent import BaseAgent
-    from .lib.subkernels.base import BaseSubkernel
+    from .lib.agent import BeakerAgent
+    from .lib.subkernels.base import BeakerSubkernel
 
 USER_RESPONSE_WAIT_TIME = 100
 
@@ -40,7 +40,7 @@ MESSAGE_STREAMS = {
 AVAILABLE_CONTEXTS = autodiscover_contexts()
 
 
-class LLMKernel(KernelProxyManager):
+class BeakerKernel(KernelProxyManager):
     implementation = "beaker-kernel"
     implementation_version = "0.1"
     banner = "Beaker Kernel"
@@ -53,7 +53,7 @@ class LLMKernel(KernelProxyManager):
 
     kernel_id: Optional[str]
     connection_file: Optional[str]
-    context: Optional[BaseContext]
+    context: Optional[BeakerContext]
     internal_executions: set[str]
     subkernel_execution_tracking: dict[str, str]
     user_responses: dict[str, str]
@@ -68,7 +68,7 @@ class LLMKernel(KernelProxyManager):
         self.magic_commands = {}
         self.internal_executions = set()
         self.subkernel_execution_tracking = {}
-        super().__init__(session_config)
+        super().__init__(session_config, session_id=f"{kernel_id}_session")
         self.register_magic_commands()
         self.add_base_intercepts()
         self.context = None
@@ -81,9 +81,6 @@ class LLMKernel(KernelProxyManager):
         """
         Adds intercepts used by the Beaker kernel
         """
-        self.server.intercept_message(
-            "shell", "kernel_info_reply", self.kernel_info_reply
-        )
         self.server.intercept_message(
             "shell", "context_info_request", self.context_info_request
         )
@@ -122,9 +119,10 @@ class LLMKernel(KernelProxyManager):
         for prefix, fn in self.magic_commands.items():
             if head == prefix:
                 self.debug(event_type="magic_word", content={"magic_word": head, "fn": fn.__name__})
-                async with handle_message(self, server, target_stream, data) as ctx:
-                    result = await fn(tail, magic_word=head, parent_header=message.header)
-                    ctx.return_val = result
+                async with handle_message(server, target_stream, data) as ctx:
+                    with parent_message_context(ctx.message):
+                        result = await fn(tail, magic_word=head, parent_header=message.header)
+                        ctx.return_val = result
                 return None
         return data
 
@@ -133,7 +131,7 @@ class LLMKernel(KernelProxyManager):
         context_name, language, context_config = cell_content.split(maxsplit=2)
         context_info = json.loads(context_config)
         self.stdout(f"Switching from context {self.context.slug} to {context_name}...", parent_header=parent_header)
-        await self.set_context(context_name=context_name, language=language, context_info=context_info, parent_header=parent_header)
+        await self.set_context(context_name=context_name, context_info=context_info, language=language, parent_header=parent_header)
 
         # Send message to trigger updating the context info
         self.send_response(
@@ -143,6 +141,8 @@ class LLMKernel(KernelProxyManager):
             parent_header=parent_header,
         )
         self.stdout(f"Context switch complete.", parent_header=parent_header)
+
+        # TODO: Return the new context info
         return None
 
     @magic("run_action")
@@ -214,11 +214,11 @@ class LLMKernel(KernelProxyManager):
         self.server.filters.remove(InterceptionFilter(socket, msg_type, func))
 
     async def send_preview(self, parent_header=None):
-        generate_preview = getattr(self.context, "generate_preview", None)
-        if generate_preview and (callable(generate_preview) or inspect.iscoroutinefunction(generate_preview)):
-            preview_payload = await generate_preview()
-            if preview_payload:
-                self.send_response("iopub", "preview", preview_payload, parent_header=parent_header)
+        if self.context.preview:
+            with execution_context("preview"):
+                preview_payload = await self.context.preview()
+                if preview_payload:
+                    self.send_response("iopub", "preview", preview_payload, parent_header=parent_header)
 
     async def update_connection_file(self, **kwargs):
         try:
@@ -240,10 +240,22 @@ class LLMKernel(KernelProxyManager):
         # Cleanup the old context, then create and setup the new context
         if self.context:
             self.context.cleanup()
-        self.context = context_cls(beaker_kernel=self, config=context_info)
-        await self.context.setup(config=context_info, parent_header=parent_header)
-        await self.send_preview(parent_header=parent_header)
+
+        config = {
+            "language": language,
+            "context_info": context_info
+        }
+        self.context = context_cls(beaker_kernel=self, config=config)
+        await self.context.setup(context_info=context_info, parent_header=parent_header)
+        subkernel = self.context.subkernel
+        kernel_setup_func = getattr(subkernel, "setup", None)
+        with execution_context(type="setup", name=context_name, parent_header=parent_header):
+            if inspect.iscoroutinefunction(kernel_setup_func):
+                await kernel_setup_func()
+            elif inspect.isfunction(kernel_setup_func) or inspect.ismethod(kernel_setup_func):
+                kernel_setup_func()
         await self.update_connection_file(context={"name": context_name, "config": context_info})
+        await self.send_preview(parent_header=parent_header)
 
     async def post_execute(self, queue, message_id, data):
         message = JupyterMessage.parse(data)
@@ -316,7 +328,7 @@ class LLMKernel(KernelProxyManager):
             "stdin",
             "input_request",
             {"prompt": query},
-            parent_header=parent_message.header,
+            parent_header=getattr(parent_message, "header", None),
             parent_identities=getattr(parent_message, "identities", None),
             msg_id=msg_id,
         )
@@ -382,11 +394,13 @@ class LLMKernel(KernelProxyManager):
             return
         if not self.context:
             raise Exception("Context has not been set")
+        setattr(self.context, "current_llm_query", request)
         try:
             # Before starting ReAct loop, replace thought handler with partial func with parent_header so we can track thoughts
-            self.context.agent.thought_handler = partial(self.handle_thoughts, parent_header=message.header)
-            self.debug("llm_query", request, parent_header=message.header)
-            result = await self.context.agent.react_async(request, react_context={"message": message})
+            # with parent_message_context(parent_message=message):
+                self.context.agent.thought_handler = partial(self.handle_thoughts, parent_header=message.header)
+                self.debug("llm_query", request, parent_header=message.header)
+                result = await self.context.agent.react_async(request, react_context={"message": message})
         except Exception as err:
             error_text = f"""LLM Error:
 {err}
@@ -428,10 +442,8 @@ class LLMKernel(KernelProxyManager):
         finally:
             # When done, put thought handler back to default to not potentially cause confused thoughts.
             self.context.agent.thought_handler = self.handle_thoughts
+            setattr(self.context, "current_llm_query", None)
 
-    async def kernel_info_reply(self, server, target_stream, data):
-        await self.send_preview(parent_header={})
-        return data
 
     @message_handler
     async def context_info_request(self, message):
@@ -439,8 +451,11 @@ class LLMKernel(KernelProxyManager):
         context_class = self.context.__class__
         context_slug = context_slugs_by_class.get(context_class, "Not Found")
         full_context_class = f"{context_class.__module__}.{context_class.__name__}"
-        context_config = getattr(self.context, "config", None)
+        context_config = getattr(self.context, "config", {}).get("context_info", None)
         context_info = self.context.get_info()
+        language_slug = self.context.subkernel.SLUG
+        subkernel_name = self.context.subkernel.KERNEL_NAME
+
         self.send_response(
             stream="iopub",
             msg_or_type="context_info_response",
@@ -449,14 +464,14 @@ class LLMKernel(KernelProxyManager):
                 "class": full_context_class,
                 "config": context_config,
                 "language": {
-                    "slug": self.context.subkernel.SLUG,
-                    "subkernel": self.context.subkernel.KERNEL_NAME,
+                    "slug": language_slug,
+                    "subkernel": subkernel_name,
                 },
                 "info": context_info,
             },
             parent_header=message.header,
         )
-        return False
+        return None
 
     @message_handler
     async def context_setup_request(self, message):
@@ -494,8 +509,11 @@ class LLMKernel(KernelProxyManager):
         parent_id = message.parent_header["msg_id"]
         self.user_responses[parent_id] = content["value"]
 
+# Provided for backwards compatibility
+LLMKernel = BeakerKernel
 
-def cleanup(kernel: LLMKernel):
+
+def cleanup(kernel: BeakerKernel):
     try:
         kernel.context.cleanup()
     except requests.exceptions.ConnectionError:
@@ -512,7 +530,7 @@ def start(connection_file):
 
     _, kernel_file = os.path.split(connection_file)
     kernel_id = os.path.basename(kernel_file)[7:-5]  # Remove 'kernel-' and '.json' from the beginning and end of the filename.
-    kernel = LLMKernel(notebook_config, kernel_id=kernel_id, connection_file=connection_file)
+    kernel = BeakerKernel(notebook_config, kernel_id=kernel_id, connection_file=connection_file)
 
     try:
         loop.start()
